
#include "StdAfx.h"

#pragma warning(disable: 4786)	// 'identifier' : identifier was truncated to 'number' characters in the debug information

#include "D3DDevice.h"
#include "GeometryData.h"
#include "FileGeometryData.h"

HRESULT D3DRenderDevice::Create()
{
	IDirect3D8 *pEnum = Direct3DCreate8(D3D_SDK_VERSION);
	if (pEnum == NULL) {
		return E_FAIL;
	}

	HRESULT hr = pEnum->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &d3ddev);

	pEnum->Release();

	if (!FAILED(hr)) {
		if (fixedAspect) {
			RECT wrect;
			RECT crect;
			GetWindowRect(hwnd, &wrect);
			GetClientRect(hwnd, &crect);

			SetWindowPos(hwnd, NULL, wrect.left, wrect.top,
				wrect.right-wrect.left + d3dpp.BackBufferWidth-crect.right,
				wrect.bottom-wrect.top + d3dpp.BackBufferHeight-crect.bottom,
				SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_NOZORDER);
		}
	}

	return hr;
}

void D3DRenderDevice::Release()
{
	if (d3ddev) {
		fontManager.ReleaseFonts();
		ReleaseTextures();
		ReleaseGeometry();

		d3ddev->Release();
		d3ddev = NULL;
	}
}

LRESULT D3DRenderDevice::WndProc(HWND phwnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool &callDefWndProc)
{
	callDefWndProc = true;
	if (hwnd != phwnd) {
		return 0;
	}
	switch (uMsg) {
		case WM_CLOSE:
			Release();
			break;

		case WM_WINDOWPOSCHANGING:
			if (fixedAspect) {
				LPWINDOWPOS lpwndpos = (LPWINDOWPOS) lParam;
				if (lpwndpos->cx > 0 && lpwndpos->cy > 0) {
					RECT wrect;
					RECT crect;
					int cx = lpwndpos->cx;
					int cy = lpwndpos->cy;
					GetWindowRect(hwnd, &wrect);
					GetClientRect(hwnd, &crect);
					int ww = wrect.right - wrect.left;
					int wh = wrect.bottom - wrect.top;
					int cw = crect.right - crect.left;
					int ch = crect.bottom - crect.top;
					if (ww > 0 && wh > 0 && cw > 0 && ch > 0) {
						if (lpwndpos->flags & SWP_NOSIZE) {
							cx = ww;
							cy = wh;
						}
						int xdif = ww - cw;
						int ydif = wh - ch;
						float ratio = GetSystemMetrics(SM_CYSCREEN) / (float) GetSystemMetrics(SM_CXSCREEN);
						int wantedy = (int) ((cx - xdif) * ratio + ydif);
						if (wantedy < cy) {
							lpwndpos->cx = cx;
							lpwndpos->cy = wantedy;
							lpwndpos->flags &= ~SWP_NOSIZE;
						} else if (wantedy > cy) {
							lpwndpos->cx = (int) ((cy - ydif) / ratio + xdif);
							lpwndpos->cy = cy;
							lpwndpos->flags &= ~SWP_NOSIZE;
						}

						callDefWndProc = false;
						return 0;
					}
				}
			}
			break;

		case WM_KEYDOWN:
			switch (wParam) {
				case VK_F11:
					{
						enum { STYLEMASK = (WS_CAPTION | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX) };

						if (swFullScreen) {
							SetWindowLong(hwnd, GWL_STYLE, GetWindowLong(hwnd, GWL_STYLE) | STYLEMASK);
							SetWindowPos(hwnd, HWND_NOTOPMOST, 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN), SWP_FRAMECHANGED);
							SetWindowPlacement(hwnd, &placement);
						} else {
							GetWindowPlacement(hwnd, &placement);
							SetWindowLong(hwnd, GWL_STYLE, GetWindowLong(hwnd, GWL_STYLE) & ~STYLEMASK);
							SetWindowPos(hwnd, HWND_TOPMOST, 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN), SWP_FRAMECHANGED);
						}
						swFullScreen = !swFullScreen;
					}

					callDefWndProc = false;
					return 0;
			}
			break;

		default:
			break;
	}

	return 0;
}

HRESULT D3DRenderDevice::CheckForLostDevice()
{
    HRESULT hr = d3ddev->Present(NULL, NULL, NULL, NULL);
    while (hr == D3DERR_DEVICELOST) {
		fontManager.ReleaseFonts();
		ReleaseTextures();
		ReleaseGeometry();

        do {
            Sleep(1000);
            hr = d3ddev->TestCooperativeLevel();
        } while (hr != D3DERR_DEVICENOTRESET);

        HDC hdc = GetDC(hwnd);
        if (hdc == NULL) {
			return D3DERR_NOTAVAILABLE;
        }

        hr = d3ddev->Reset(&d3dpp);

        ReleaseDC(hwnd, hdc);

		return D3DERR_DEVICELOST;
    }

    return D3D_OK;
}

// Textures.

IDirect3DTexture8 *D3DRenderDevice::GetTexture(const std::string &texturename) const
{
	std::map<std::string, IDirect3DTexture8*>::const_iterator it = m_TextureMap.find(texturename);
	if (it == m_TextureMap.end()) {
		// Load the texture now.
		IDirect3DTexture8 *texture = NULL;
		D3DXCreateTextureFromFileA(d3ddev, texturename.c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("..\\" + texturename).c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("..\\..\\" + texturename).c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("..\\..\\..\\" + texturename).c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("Media\\" + texturename).c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("..\\Media\\" + texturename).c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("..\\..\\Media\\" + texturename).c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("..\\..\\..\\Media\\" + texturename).c_str(), &texture);
		if (!texture) D3DXCreateTextureFromFileA(d3ddev, ("c:\\NewAquaData\\" + texturename).c_str(), &texture);
		m_TextureMap[texturename] = texture;
		return texture;
	} else {
		return it->second;
	}
}

void D3DRenderDevice::ReleaseTextures()
{
	std::map<std::string, IDirect3DTexture8*>::iterator it;
	for (it = m_TextureMap.begin(); it != m_TextureMap.end(); ++it) {
		if (it->second) {
			it->second->Release();
		}
	}
	m_TextureMap.clear();
}

// Text fonts.

void D3DRenderDevice::AddGeometry(const std::string &geomname, GeometryData *geom) const
{
	if (!geom) {
		geom = new FileGeometryData(geomname);
	}
	std::map<std::string, GeometryData *>::iterator it = m_GeometryMap.find(geomname);
	if (it == m_GeometryMap.end()) {
		m_GeometryMap[geomname] = geom;
	} else {
		delete it->second;
		it->second = geom;
	}
}

GeometryData *D3DRenderDevice::GetGeometry(const std::string &geomname) const
{
	std::map<std::string, GeometryData *>::const_iterator it = m_GeometryMap.find(geomname);
	if (it == m_GeometryMap.end()) {
		return NULL;
	} else {
		return it->second;
	}
}

void D3DRenderDevice::ReleaseGeometry()
{
	std::map<std::string, GeometryData *>::iterator it;
	for (it = m_GeometryMap.begin(); it != m_GeometryMap.end(); ++it) {
		delete it->second;
	}
	m_GeometryMap.clear();
}
