
#ifndef D3DDevice_H_INCLUDED
#define D3DDevice_H_INCLUDED

#pragma once

#include "FontManager.h"

#include <d3d8.h>
#include <d3dx8.h>

#include <string>
#include <map>

class GeometryData;

template < class T >
class PropertySet {
private:
	std::map<std::string, T> properties;
public:
	const T &get(const std::string &name) const {
		std::map<std::string, T>::const_iterator it = properties.find(name);
		if (it != properties.end()) {
			return it->second;
		} else {
			static const T def = T();
			return def;
		}
	}
	void set(const std::string &name, const T &t) {
		properties[name] = t;
	}
};

struct Properties3D {
	PropertySet<unsigned int>	Int;
	PropertySet<float		>	Float;
	PropertySet<D3DXCOLOR	>	Color;
	PropertySet<D3DXVECTOR3	>	Vector3;
	PropertySet<D3DXMATRIX	>	Matrix;
};

// Property names:
const std::string PropProjectionMatrix	= "Projection";
const std::string PropViewMatrix		= "View";
const std::string PropWorldMatrix		= "World";
const std::string PropCameraPos			= "Camera";
const std::string PropTopFogColor		= "FogTop";
const std::string PropBottomFogColor	= "FogBottom";
const std::string PropFogNear			= "FogNear";
const std::string PropFogFar			= "FogFar";
const std::string PropBlamFogDir		= "FogBlamDir";
const std::string PropBlamFogWidth		= "FogBlamWidth";

class D3DRenderDevice {
public:
	HWND					hwnd;
	bool					fixedAspect;
	D3DPRESENT_PARAMETERS	d3dpp;
	IDirect3DDevice8 *		d3ddev;
	Properties3D			prop;
	bool					swFullScreen;
	WINDOWPLACEMENT			placement;
	FontManager				fontManager;

	IDirect3DDevice8 *operator->() const { return d3ddev; }
	operator IDirect3DDevice8 *() const { return d3ddev; }
	operator const D3DPRESENT_PARAMETERS &() const { return d3dpp; }
	operator D3DPRESENT_PARAMETERS &() { return d3dpp; }
	operator HWND() const { return hwnd; }
	operator bool() const { return d3ddev != 0; }

	HRESULT Create();
	void Release();

	LRESULT WndProc(HWND phwnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool &callDefWndProc);

	HRESULT CheckForLostDevice();

	// Textures.
	mutable std::map<std::string, IDirect3DTexture8*> m_TextureMap;
	IDirect3DTexture8 *GetTexture(const std::string &texturename) const;
	void ReleaseTextures();

	// Text fonts.
	ID3DXFont *GetFont(const std::string &name, int size, bool bold, bool italic, bool underline) const {
		return fontManager.GetFont(*this, GetFontName(name, size, bold, italic, underline));
	}
	ID3DXFont *GetFont(const FontName &fontname) const {
		return fontManager.GetFont(*this, fontname);
	}

	// Geometry.
	mutable std::map<std::string, GeometryData*> m_GeometryMap;
	void AddGeometry(const std::string &geomname, GeometryData *geom = 0) const;
	GeometryData *GetGeometry(const std::string &geomname) const;
	void ReleaseGeometry();

	D3DRenderDevice():
		hwnd(0),
		fixedAspect(false),
		d3ddev(0),
		swFullScreen(false)
	{
	}
};

#endif
