
#include "StdAfx.h"

#pragma warning(disable: 4786)	// 'identifier' : identifier was truncated to 'number' characters in the debug information

#include "VolumetricFog.h"

FogVolumeDistanceFlat::FogVolumeDistanceFlat(
		float			minDist,
		float			maxDist,
		const D3DCOLOR &color):
	m_MinDist(minDist),
	m_MaxDist(maxDist),
	m_Color(color)
{
	if (m_MinDist < 0) {
		m_MinDist = 0;
	}
	if (m_MaxDist < m_MinDist) {
		const float temp = m_MinDist;
		m_MinDist = m_MaxDist;
		m_MaxDist = temp;
	}
	m_Color.a = 1.F / (m_MaxDist - m_MinDist);
	m_Color.r *= m_Color.a;
	m_Color.g *= m_Color.a;
	m_Color.b *= m_Color.a;
}

void FogVolumeDistanceFlat::GetFogValues(
	FogChangeValue *&	changeValueBuffer,
	D3DXCOLOR &			initialColor,
	const D3DXVECTOR3 &	cameraPos,
	D3DXVECTOR3 &		vertexPos,
	float &				distance)
{
	float minDist = m_MinDist;
	float maxDist = m_MaxDist;
	D3DXCOLOR color = m_Color;

	CalculateValues(cameraPos, vertexPos, minDist, maxDist, color);

	D3DXVECTOR3 distanceVector = vertexPos - cameraPos;
	distance = D3DXVec3Length(&distanceVector);

	if (distance > maxDist) {
		// Optimization:
		// If the vertex is behind the far end of the fog,
		// then we bring it closer to lie on it.
		distanceVector *= maxDist / distance;
		vertexPos	= cameraPos + distanceVector;
		distance	= maxDist;

		// Note that the value of "t" would be always 1, so we don't need to add it.
	}

	if (minDist == 0.F) {
		// Optimization:
		// if the near point of the fog is AT the observer,
		// we just add the color and density to the initial values.

		initialColor += color;
	} else if (distance > minDist) {
		// Optimization:
		// We add the near fog point only if it's closer than the vertex.
		// i.e. if minT would've ended up greater than 1.

		float minT = minDist / distance;

		changeValueBuffer[0].t = minT;
		changeValueBuffer[0].color = &color;
		++changeValueBuffer;
	}
}

FogVolumeDistanceVGrad::FogVolumeDistanceVGrad(
		float			minDist,
		float			maxDist,
		const D3DCOLOR &hicolor,
		const D3DCOLOR &locolor):
	FogVolumeDistanceFlat(minDist, maxDist, locolor),
	m_HiColor(hicolor),
	m_LoColor(locolor)
{
}

void FogVolumeDistanceVGrad::CalculateValues(const D3DXVECTOR3 &cameraPos, const D3DXVECTOR3 &vertexPos, float &minDist, float &maxDist, D3DXCOLOR &color)
{
	D3DXVECTOR3 vdir = vertexPos - cameraPos;
	float v = (vdir.y / D3DXVec3Length(&vdir) + 1.F) * .5F;
//	v = 3*v*v - 2*v*v*v;

	color.r = (m_LoColor.r + (m_HiColor.r - m_LoColor.r) * v) * color.a;
	color.g = (m_LoColor.g + (m_HiColor.g - m_LoColor.g) * v) * color.a;
	color.b = (m_LoColor.b + (m_HiColor.b - m_LoColor.b) * v) * color.a;

	FogVolumeDistanceFlat::CalculateValues(cameraPos, vertexPos, minDist, maxDist, color);
}

FogVolumeDistanceHGrad::FogVolumeDistanceHGrad(
		float			minDist,
		float			maxDist,
		const D3DCOLOR &hicolor,
		const D3DCOLOR &locolor):
	FogVolumeDistanceFlat(minDist, maxDist, locolor),
	m_HiColor(hicolor),
	m_LoColor(locolor)
{
}

void FogVolumeDistanceHGrad::CalculateValues(const D3DXVECTOR3 &cameraPos, const D3DXVECTOR3 &vertexPos, float &minDist, float &maxDist, D3DXCOLOR &color)
{
	D3DXVECTOR3 vdir = vertexPos - cameraPos;
	float v = (vdir.x / D3DXVec3Length(&vdir) + 1.F) * .5F;
//	v = 3*v*v - 2*v*v*v;

	color.r = (m_LoColor.r + (m_HiColor.r - m_LoColor.r) * v) * color.a;
	color.g = (m_LoColor.g + (m_HiColor.g - m_LoColor.g) * v) * color.a;
	color.b = (m_LoColor.b + (m_HiColor.b - m_LoColor.b) * v) * color.a;

	FogVolumeDistanceFlat::CalculateValues(cameraPos, vertexPos, minDist, maxDist, color);
}

FogVolumeDistanceVGradBlam::FogVolumeDistanceVGradBlam(
		float			minDist,
		float			maxDist,
		const D3DCOLOR &hicolor,
		const D3DCOLOR &locolor):
	FogVolumeDistanceVGrad(minDist, maxDist, hicolor, locolor),
	m_BlamWidth(0.F)
{
}

void FogVolumeDistanceVGradBlam::CalculateValues(const D3DXVECTOR3 &cameraPos, const D3DXVECTOR3 &vertexPos, float &minDist, float &maxDist, D3DXCOLOR &color)
{
	FogVolumeDistanceVGrad::CalculateValues(cameraPos, vertexPos, minDist, maxDist, color);
	if (m_BlamWidth >= 0.F) {
		D3DXVECTOR3 vdir = vertexPos - cameraPos;
		D3DXVec3Normalize(&vdir, &vdir);
		D3DXVECTOR3 bdir = m_BlamPos - cameraPos;
		float blamdist = D3DXVec3Length(&bdir);
		bdir /= blamdist;
		float blamv = D3DXVec3Dot(&vdir, &bdir);
		blamv = (blamv + 1.F) * .5F;
		blamv = 1.F - blamv;
		blamv *= blamdist;
		if (blamv < m_BlamWidth) {
			blamv = (1.F - blamv / m_BlamWidth) / maxDist;
			color.r += m_BlamColor.r * blamv;
			color.g += m_BlamColor.g * blamv;
			color.b += m_BlamColor.b * blamv;
			color.a += blamv;
		}
	}
}

inline D3DXVECTOR3 ComponentProd(const D3DXVECTOR3 &v1, const D3DXVECTOR3 &v2)
{
	return D3DXVECTOR3(v1.x * v2.x, v1.y * v2.y, v1.z * v2.z);
}

FogEllipsoid::FogEllipsoid(
		const D3DXVECTOR3 &	pos,
		const D3DXVECTOR3 &	radii,
		const D3DCOLOR &	color,
		float				density):
	m_Pos(pos),
	m_Scale(1.F / radii.x, 1.F / radii.y, 1.F / radii.z),
	m_Color(color)
{
	m_Color.a = density;
	m_Color.r *= m_Color.a;
	m_Color.g *= m_Color.a;
	m_Color.b *= m_Color.a;
	m_NegColor = -m_Color;
}

void FogEllipsoid::GetFogValues(
	FogChangeValue *&	changeValueBuffer,
	D3DXCOLOR &			initialColor,
	const D3DXVECTOR3 &	cameraPos,
	const D3DXVECTOR3 &	vertexPos,
	float				distance)
{
	const D3DXVECTOR3 adjCameraPos = ComponentProd((cameraPos - m_Pos), m_Scale);
	const D3DXVECTOR3 adjVertexPos = ComponentProd((vertexPos - m_Pos), m_Scale);
	const D3DXVECTOR3 adjDistance = adjVertexPos - adjCameraPos;

	const float OD = D3DXVec3Dot(&adjDistance, &adjCameraPos);
	const float D2 = D3DXVec3Dot(&adjDistance, &adjDistance);
	const float O2 = D3DXVec3Dot(&adjCameraPos, &adjCameraPos);

	const float radix = OD*OD - D2*(O2 - 1);
	if (radix <= 0) {
		// Early out when the visibility ray doesn't intersect the fog volume.
		return;
	}

	const float sradix = float(sqrt(radix));

	float t1 = (-OD - sradix) / D2;
	float t2 = (-OD + sradix) / D2;

	if (t2 <= 0.F || t1 >= 1.F) {
		// Optimization: early out when the fog is completely outside of the visible range.
		// That is, either fully behind the observer, or fully behind the vertex.
		return;
	}

	if (t1 <= 0.F) {
		initialColor += m_Color;
	} else {
		changeValueBuffer[0].t = t1;
		changeValueBuffer[0].color = &m_Color;
		++changeValueBuffer;
	}

	if (t2 >= 1.F) {
		// Optimization: if the fog ends behind the vertex, just ignore that point.
	} else {
		changeValueBuffer[0].t = t2;
		changeValueBuffer[0].color = &m_NegColor;
		++changeValueBuffer;
	}
}

D3DXCOLOR ApplyFogsToVertex(
	FogVolumeDistance *	distanceFog,
	FogVolume **		volumeList,
	unsigned int		numVolumes,
	const D3DXVECTOR3 &	vertexPos,
	const D3DXVECTOR3 &	cameraPos)
{
	D3DXCOLOR initialColor(0.F, 0.F, 0.F, 0.F);

	D3DXVECTOR3 adjVertexPos = vertexPos;
	float distance;

	FogChangeValue changeValueBuffer[32];
	FogChangeValue *changeValueLast = changeValueBuffer;

	if (distanceFog) {
		distanceFog->GetFogValues(changeValueLast, initialColor, cameraPos, adjVertexPos, distance);
	} else {
		distance = D3DXVec3Length(&(vertexPos - cameraPos));
	}

	unsigned int i;
	for (i = 0; i < numVolumes; ++i) {
		volumeList[i]->GetFogValues(changeValueLast, initialColor, cameraPos, adjVertexPos, distance);
	}

	const unsigned int numChangeValues = changeValueLast - changeValueBuffer;

	for (i = 1; i < numChangeValues; ++i) {
		if (changeValueBuffer[i-1].t > changeValueBuffer[i].t) {
			FogChangeValue temp = changeValueBuffer[i];
			changeValueBuffer[i] = changeValueBuffer[i-1];
			unsigned int j;
			for (j = i-1; j > 0 && changeValueBuffer[j-1].t > temp.t; --j) {
				changeValueBuffer[j] = changeValueBuffer[j-1];
			}
			changeValueBuffer[j] = temp;
		}
	}

	D3DXCOLOR fog(0, 0, 0, 1);
	D3DXCOLOR accumFog = initialColor;
	float lastT = 0.F;

	for (i = 0; i < numChangeValues; ++i) {
		const FogChangeValue &changeValue = changeValueBuffer[i];
		if (accumFog.a != 0.F && changeValue.t != lastT) {
			float volFog	= (changeValue.t - lastT) * distance * accumFog.a;
//			volFog *= volFog;
			volFog = 1.F - volFog;
			const float ooaccumFogA = 1.F / accumFog.a;
			const float volFogR	= accumFog.r * ooaccumFogA;
			const float volFogG	= accumFog.g * ooaccumFogA;
			const float volFogB	= accumFog.b * ooaccumFogA;
			fog.r = volFogR * fog.a + fog.r * (1.F - fog.a);
			fog.g = volFogG * fog.a + fog.g * (1.F - fog.a);
			fog.b = volFogB * fog.a + fog.b * (1.F - fog.a);
			fog.a = volFog - 1.F + fog.a;
			if (fog.a <= 0.F) {
				fog.a = 0.F;
				return fog;
			}
		}
		accumFog += *changeValue.color;
		lastT = changeValue.t;
	}

	if (accumFog.a != 0.F && lastT < 1.F) {
		float volFog	= (1.F - lastT) * distance * accumFog.a;
//		volFog *= volFog;
		volFog = 1.F - volFog;
		const float ooaccumFogA = 1.F / accumFog.a;
		const float volFogR	= accumFog.r * ooaccumFogA;
		const float volFogG	= accumFog.g * ooaccumFogA;
		const float volFogB	= accumFog.b * ooaccumFogA;
		fog.r = volFogR * fog.a + fog.r * (1.F - fog.a);
		fog.g = volFogG * fog.a + fog.g * (1.F - fog.a);
		fog.b = volFogB * fog.a + fog.b * (1.F - fog.a);
		fog.a = volFog - 1.F + fog.a;
		if (fog.a <= 0.F) {
			fog.a = 0.F;
		}
	}

	return fog;
}

